#!/bin/bash
# Run MYULA and the serial version of the proposed Split-Gibbs Sampler (SGS)
# over all datasets considered in Section 6.3.
#
# author: pthouvenin (pierre-antoine.thouvenin@centralelille.fr)
#
# reference: P.-A. Thouvenin, A. Repetti, P. Chainais - **A distributed Gibbs
# Sampler with Hypergraph Structure for High-Dimensional Inverse Problems**,
# [arxiv preprint 2210.02341](http://arxiv.org/abs/2210.02341), October 2022.

# * Notes
#
# ! Parameter passed as explicit command line arguments override those from
# ! the configuration .json file
#
# - Results from serial samplers are saved in a folder named with the convention
#
# "<dpath>/<image_name>_M<max_intensity>_ks<kernel_size>_seed<seed>", with:
#
# <dpath>           relative path to the synthetic datasets
# <image_name>      name of the ground truth image from which the dataset has
#                   been generated
# <max_intensity>   max. intensity of the ground truth image used
# <kernel_size>     size of the convolution kernel for the dataset considered
# <seed>            seed used to initialize the random number generator
#                   underlying the sampler
#
# ! This script can only be run once the synthetic data have been generated
# ! (see "generate_data.sh")
#
# ! BEWARE: running all the simulations below requires a large amount of space
# ! on the hard drive (~210 GB). All samples generated by the serial samplers
# ! are saved to disk in the form of multiple checkpoint files, from which the
# ! samplers can be restarted to obtain a longer chain)


# relative path where the datasets will be retrieved
dpath=data

# root of the path where the results will be saved
rpath=results

# name of configuration files for the current experiment
# (delete some configuration names to run only a subset of the experiments)
declare -a experiments_configs=("configs/peppers_M20_ks7.json" "configs/peppers_M20_ks15.json" "configs/peppers_M30_ks7.json" "configs/peppers_M30_ks15.json", "configs/house_M20_ks3.json" "configs/house_M20_ks7.json" "configs/house_M30_ks3.json" "configs/house_M30_ks7.json")
declare -a samplers=("psgla" "myula")

# * debug test
# declare -a experiments_configs=("configs/tests/test_config_spmd.json")
# declare -a samplers=("psgla" "myula")

restart=-1              # iteration index identifying the checkpoint from which
                        # to restart the sampler
loadflag=''             # '--load' if loading checkpoint from disk, '' otherwise
rfile=final_results
results_text_filename=sampling_quality_metrics


# loop over datasets
for configfile in "${experiments_configs[@]}"; do

    # loop over number of MPI cores
    for sampler in "${samplers[@]}"; do
        echo Sampling quality experiment: ${configfile}, sampler=${sampler} ncores=1;

        # run serial SGS sampler
        python ../../src/dsgs/main_serial_poisson_deconvolution.py \
        --config_file=${configfile} \
        --sampler=${sampler} \
        --dpath=${dpath} --rpath=${rpath} \
        --verbose --restart=${restart} ${loadflag}

        # compute and append estimation quality results to a single .txt file
        mpiexec -n 1 python -m mpi4py ../../src/dsgs/main_metrics_serial.py \
        --config_file=${configfile} \
        --sampler=${sampler} \
        --results_text_filename=${results_text_filename} \
        --dpath=${dpath} --rpath=${rpath} --rfile=${rfile}
    done
done
