"""Helper functions to add a Metropolis-Hastings step :cite:p:`Jones2014` to
other accept-reject candidate samples.
"""

# author: pthouvenin (pierre-antoine.thouvenin@centralelille.fr)
#
# reference: P.-A. Thouvenin, A. Repetti, P. Chainais - **A distributed Gibbs
# Sampler with Hypergraph Structure for High-Dimensional Inverse Problems**,
# [arxiv preprint 2210.02341](http://arxiv.org/abs/2210.02341), October 2022.

import numpy as np

from dsgs.samplers.transition_kernels.base_kernel import BaseMCMCKernel


class MetropolisHastings(BaseMCMCKernel):
    """Generic implementation of a Metropolis-Hastings transition kernel."""

    def __init__(self):
        super(MetropolisHastings, self).__init__(params=None)

    def mh_ratio(px, py, p_xy, pyx):
        r"""Compute the Hastings ratio defined by

        .. math::
            r(x, y) = \frac{\pi(y)}{\pi(x)} \frac{g(y, x)}{g(x, y)},

        where :math:`\pi` is the (unormalised) density to be sampled, and
        :math:`g(x, y)` is the density of the Markov transition kernel from
        state :math:`x` to state :math:`y`. The acceptance ratio is given by

        .. math::
            a(x, y) = 1 \wedge r(x, y).

        Parameters
        ----------
        px : float
            Value :math:`\pi(x)` of the target density evaluated in :math:`x`.
        py : float
            Value :math:`\pi(y)` of the target density evaluated in :math:`y`.
        p_xy : float
            Evaluation of the transition density :math:`g(x, y)`.
        pyx : float
            Evaluation of the transition density :math:`g(y, x)`.

        Returns
        -------
        float
            Hastings ratio.
        """
        return (py * pyx) / (px * p_xy)

    def mh_log_ratio(log_px, log_py, log_pxy, log_pyx):
        r"""Compute the log Hasting ratio :math:`\log r(x, y)`.

        Parameters
        ----------
        log_px : float
            Value :math:`\log \pi(x)`, log of the target density evaluated in
            :math:`x`.
        log_py : float
            Value :math:`\log \pi(y)`, log of the target density evaluated in
            :math:`y`.
        log_pxy : float
            Value of :math:`\log g(x, y)`.
        log_pyx : float
            Value of :math:`\log g(y, x)`.

        Returns
        -------
        float
            Log of the Hastings ratio.
        """
        return log_py + log_pyx - (log_px + log_pxy)

    def make_transition(self, current_state, new_state, px, py, p_xy, pyx, rng):
        """Implementation of the Metropolis-Hastings step (acceptance
        probability computed in linear scale).

        Parameters
        ----------
        current_state : numpy.ndarray
            Current state.
        new_state : numpy.ndarray
            Proposed state.
        px : float
            Probability of the current state (for the proposal mechanism).
        py : float
            Probability of the proposed state (for the proposal mechanism).
        p_xy : float
            Transition probability (moving from the current to the proposed
            state).
        pyx : float
            Transition probability (moving from the proposed to the current
            state).
        rng : numpy.random.Generator
            Random number generator.

        Returns
        -------
        numpy.ndarray
            State adopted for the next iteration.
        """
        r = self.mh_ratio(px, py, p_xy, pyx)

        if r >= 1 or rng.uniform() < r:
            return new_state
        else:
            return current_state

    def make_transition_log(
        self, current_state, new_state, log_px, log_py, log_p_xy, log_pyx, rng
    ):
        """Implementation of the Metropolis-Hastings step (acceptance
        probability computed in log scale).

        Parameters
        ----------
        current_state : numpy.ndarray
            Current state.
        new_state : numpy.ndarray
            Proposed state.
        log_px : float
            Log probability of the current state (for the proposal mechanism).
        log_py : float
            Log probability of the proposed state (for the proposal mechanism).
        log_p_xy : float
            Log of the transition probability (moving from the current to the
            proposed state).
        log_pyx : float
            Log of the transition probability (moving from the proposed to the
            current state).
        rng : numpy.random.Generator
            Random number generator.

        Returns
        -------
        numpy.ndarray
            State adopted for the next iteration.
        """
        log_r = self.mh_log_ratio(log_px, log_py, log_p_xy, log_pyx)

        if log_r >= 0 or rng.uniform() < np.exp(log_r):
            return new_state
        else:
            return current_state


class MetropolisedKernel(MetropolisHastings):
    """Generic implementation of a Metropolis-Hastings transition kernel, where
    the proposal distribution is generated by another kernel."""

    def __init__(self, proposal_kernel: BaseMCMCKernel):
        # ! see if valid, and if changes in proposal_kernel used for the
        # ! initialization are tracked within the outer object
        self.proposal_kernel = proposal_kernel
        super(MetropolisHastings, self).__init__(params=None)

    def make_transition(self, current_state, new_state, px, py, p_xy, pyx, rng):
        """Implementation of the Metropolis-Hastings step (acceptance
        probability computed in linear scale).

        Parameters
        ----------
        current_state : numpy.ndarray
            Current state.
        new_state : numpy.ndarray
            Proposed state.
        px : float
            Probability of the current state (for the proposal mechanism).
        py : float
            Probability of the proposed state (for the proposal mechanism).
        p_xy : float
            Transition probability (moving from the current to the proposed
            state).
        pyx : float
            Transition probability (moving from the proposed to the current
            state).
        rng : numpy.random.Generator
            Random number generator.

        Returns
        -------
        numpy.ndarray
            State adopted for the next iteration.
        """
        r = self.mh_ratio(px, py, p_xy, pyx)

        if r >= 1 or rng.uniform() < r:
            return new_state
        else:
            return current_state

    def make_transition_log(
        self, current_state, new_state, log_px, log_py, log_p_xy, log_pyx, rng
    ):
        """Implementation of the Metropolis-Hastings step (acceptance
        probability computed in log scale).

        Parameters
        ----------
        current_state : numpy.ndarray
            Current state.
        new_state : numpy.ndarray
            Proposed state.
        log_px : float
            Log probability of the current state (for the proposal mechanism).
        log_py : float
            Log probability of the proposed state (for the proposal mechanism).
        log_p_xy : float
            Log of the transition probability (moving from the current to the
            proposed state).
        log_pyx : float
            Log of the transition probability (moving from the proposed to the
            current state).
        rng : numpy.random.Generator
            Random number generator.

        Returns
        -------
        numpy.ndarray
            State adopted for the next iteration.
        """
        log_r = self.mh_log_ratio(log_px, log_py, log_p_xy, log_pyx)

        if log_r >= 0 or rng.uniform() < np.exp(log_r):
            return new_state
        else:
            return current_state
